import asyncio
import concurrent.futures
import os
import random
import shlex
import string
import subprocess
import time
from collections import deque
from io import StringIO
from typing import Optional, AsyncIterable, Any

# noinspection PyPackageRequirements
import discord
import nacl.hash
import youtube_dl
from ptvp35 import Db, KVJson
from v6d0auth.config import root
from v6d0auth.run_app import start_app
from v6d1tokens.client import request_token
from v6d2ctx.context import Context, of, at, escape, Implicit, monitor, Benchmark, Explicit

from v6d3music.app import get_app
from v6d3music.config import prefix

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)
token = loop.run_until_complete(request_token('music'))
client = discord.Client(
    intents=discord.Intents(
        members=True,
        guilds=True,
        bans=True,
        emojis=True,
        invites=True,
        voice_states=True,
        guild_messages=True,
        reactions=True
    ),
)
myroot = root / 'v6d3music'
myroot.mkdir(exist_ok=True)
volume_db = Db(myroot / 'volume.db', kvrequest_type=KVJson)
queue_db = Db(myroot / 'queue.db', kvrequest_type=KVJson)
cache_db = Db(myroot / 'cache.db', kvrequest_type=KVJson)
cache_root = myroot / 'cache'
cache_root.mkdir(exist_ok=True)
cache_locks: dict[str, asyncio.Lock] = {}

vcs_restored = False


async def restore_vcs():
    global vcs_restored
    vcs: list[tuple[int, int, bool]] = queue_db.get('vcs', [])
    try:
        for vcgid, vccid, vc_is_paused in vcs:
            try:
                guild: discord.Guild = await client.fetch_guild(vcgid)
                async with lock_for(guild):
                    channels = await guild.fetch_channels()
                    channel: discord.VoiceChannel
                    channel, = [ch for ch in channels if ch.id == vccid]
                    vp: discord.VoiceProtocol = await channel.connect()
                    assert isinstance(vp, discord.VoiceClient)
                    vc = vp
                    await main_for_raw_vc(vc, create=True)
                    if vc_is_paused:
                        vc.pause()
            except Exception as e:
                print(f'vc {vcgid} {vccid} {vc_is_paused} failed', e)
            else:
                print(f'vc restored {vcgid} {vccid}')
    finally:
        vcs_restored = True


@client.event
async def on_ready():
    print("ready")
    await client.change_presence(activity=discord.Game(
        name='феноменально',
    ))
    if not vcs_restored:
        await restore_vcs()


async def handle_command(ctx: Context, name: str, args: list[str]) -> None:
    await of('commands', name)(ctx, args)


@at('commands', 'help')
async def help_(ctx: Context, args: list[str]) -> None:
    match args:
        case []:
            await ctx.reply('music bot')
        case [name]:
            await ctx.reply(f'help for {name}: `{name} help`')


locks: dict[discord.Guild, asyncio.Lock] = {}


def lock_for(guild: discord.Guild) -> asyncio.Lock:
    if guild is None:
        raise Explicit('not in a guild')
    if guild in locks:
        return locks[guild]
    else:
        return locks.setdefault(guild, asyncio.Lock())


def cache_lock_for(hurl: str) -> asyncio.Lock:
    if hurl in cache_locks:
        return cache_locks[hurl]
    else:
        return cache_locks.setdefault(hurl, asyncio.Lock())


class YTAudio(discord.AudioSource):
    source: discord.FFmpegPCMAudio

    def __init__(
            self,
            url: str,
            origin: str,
            description: str,
            options: Optional[str],
            rby: discord.Member,
            already_read: int,
    ):
        self.url = url
        self.origin = origin
        self.description = description
        self.options = options
        self.rby = rby
        self.already_read = already_read
        self.loaded = False
        self.regenerating = False
        self.set_source()

    def set_source(self):
        self.source = discord.FFmpegPCMAudio(
            self.url,
            options=self.options,
            before_options=self.before_options()
        )

    def before_options(self):
        before_options = ''
        if 'https' in self.url:
            before_options += (
                '-reconnect 1 -reconnect_at_eof 0 -reconnect_streamed 1 -reconnect_delay_max 10 -copy_unknown'
            )
        if self.already_read:
            before_options += f' -ss {self.already_read * discord.opus.Encoder.FRAME_LENGTH / 1000}'
        return before_options

    def read(self) -> bytes:
        if self.regenerating:
            return FILL
        self.already_read += 1
        ret: bytes = self.source.read()
        if ret:
            self.loaded = True
        elif not self.loaded:
            if random.random() > .1:
                self.regenerating = True
                loop.create_task(self.regenerate())
                return FILL
            else:
                print(f'dropped {self.origin}')
        return ret

    def cleanup(self):
        self.source.cleanup()

    def can_be_skipped_by(self, member: discord.Member) -> bool:
        permissions: discord.Permissions = member.guild_permissions
        if permissions.administrator:
            return True
        elif permissions.manage_permissions:
            return True
        elif permissions.manage_guild:
            return True
        elif permissions.manage_channels:
            return True
        elif permissions.manage_messages:
            return True
        else:
            return self.rby == member

    def hybernate(self):
        return {
            'url': self.url,
            'origin': self.origin,
            'description': self.description,
            'options': self.options,
            'rby': self.rby.id,
            'already_read': self.already_read,
        }

    @classmethod
    async def respawn(cls, guild: discord.Guild, respawn) -> 'YTAudio':
        return YTAudio(
            respawn['url'],
            respawn['origin'],
            respawn['description'],
            respawn['options'],
            guild.get_member(respawn['rby']) or await guild.fetch_member(respawn['rby']),
            respawn['already_read']
        )

    async def regenerate(self):
        try:
            print(f'regenerating {self.origin}')
            self.url = await real_url(self.origin, True)
            self.source.cleanup()
            self.set_source()
            print(f'regenerated {self.origin}')
        finally:
            self.regenerating = False


FILL = b'\x00' * discord.opus.Encoder.FRAME_SIZE


class QueueAudio(discord.AudioSource):
    def __init__(self, guild: discord.Guild, respawned: list[YTAudio]):
        self.queue: deque[YTAudio] = deque()
        self.queue.extend(respawned)
        self.guild = guild

    @staticmethod
    async def respawned(guild: discord.Guild) -> list[YTAudio]:
        respawned = []
        try:
            for audio_respawn in queue_db.get(guild.id, []):
                try:
                    respawned.append(await YTAudio.respawn(guild, audio_respawn))
                except Exception as e:
                    print('audio respawn failed', e)
                    raise
        except Exception as e:
            print('queue respawn failed', e)
        return respawned

    @classmethod
    async def create(cls, guild: discord.Guild):
        return cls(guild, await QueueAudio.respawned(guild))

    async def save(self):
        hybernated = []
        for audio in list(self.queue):
            await asyncio.sleep(0.01)
            hybernated.append(audio.hybernate())
        queue_db.set_nowait(self.guild.id, hybernated)

    def append(self, audio: YTAudio):
        self.queue.append(audio)

    def read(self) -> bytes:
        if not self.queue:
            return FILL
        audio = self.queue[0]
        frame = audio.read()
        if len(frame) != discord.opus.Encoder.FRAME_SIZE:
            self.queue.popleft().cleanup()
            frame = FILL
        return frame

    def skip_at(self, pos: int, member: discord.Member) -> bool:
        if pos < len(self.queue):
            audio = self.queue[pos]
            if audio.can_be_skipped_by(member):
                self.queue.remove(audio)
                audio.cleanup()
                return True
        return False

    def skip_audio(self, audio: YTAudio, member: discord.Member) -> bool:
        if audio in self.queue:
            if audio.can_be_skipped_by(member):
                self.queue.remove(audio)
                audio.cleanup()
                return True
        return False

    def clear(self, member: discord.Member) -> None:
        permissions: discord.Permissions = member.guild_permissions
        if permissions.administrator:
            self.cleanup()
        else:
            raise Explicit('not an administrator')

    def swap(self, member: discord.Member, a: int, b: int) -> None:
        permissions: discord.Permissions = member.guild_permissions
        if permissions.administrator:
            if max(a, b) >= len(self.queue):
                return
            self.queue[a], self.queue[b] = self.queue[b], self.queue[a]
        else:
            raise Explicit('not an administrator')

    def move(self, member: discord.Member, a: int, b: int) -> None:
        permissions: discord.Permissions = member.guild_permissions
        if permissions.administrator:
            if max(a, b) >= len(self.queue):
                return
            audio = self.queue[a]
            self.queue.remove(audio)
            self.queue.insert(b, audio)
        else:
            raise Explicit('not an administrator')

    def format(self) -> str:
        stream = StringIO()
        for i, audio in enumerate(list(self.queue)):
            stream.write(f'`[{i}]` {audio.description}\n')
        return stream.getvalue()

    def cleanup(self):
        self.queue.clear()
        for audio in self.queue:
            try:
                audio.cleanup()
            except ValueError:
                pass


class MainAudio(discord.PCMVolumeTransformer):
    def __init__(self, queue: QueueAudio, volume: float):
        self.queue = queue
        super().__init__(self.queue, volume=volume)

    async def set(self, volume: float, member: discord.Member):
        if volume < 0.01:
            raise Explicit('volume too small')
        if volume > 1:
            raise Explicit('volume too big')
        permissions: discord.Permissions = member.guild_permissions
        if not permissions.administrator:
            raise Explicit('not an administrator')
        self.volume = volume
        await volume_db.set(member.guild.id, volume)


def extract(params: dict, url: str, kwargs: dict):
    extracted = youtube_dl.YoutubeDL(params=params).extract_info(url, **kwargs)
    if 'entries' in extracted:
        extracted['entries'] = list(extracted['entries'])
    return extracted


def bytes_hash(b: bytes) -> str:
    return nacl.hash.sha256(b).decode()


def recursive_hash(obj) -> str:
    if isinstance(obj, str):
        return bytes_hash(obj.encode())
    elif isinstance(obj, tuple) or isinstance(obj, list):
        return recursive_hash(';'.join(map(recursive_hash, obj)))
    elif isinstance(obj, dict):
        return recursive_hash([*obj.items()])
    else:
        raise TypeError


async def aextract(params: dict, url: str, **kwargs):
    with Benchmark('AEX'):
        with concurrent.futures.ProcessPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool,
                extract,
                params,
                url,
                kwargs
            )


async def cache_url(hurl: str, rurl: str, override: bool) -> None:
    async with cache_lock_for(hurl):
        if not override and cache_db.get(f'url:{hurl}', None) is not None:
            return
        cachable: bool = cache_db.get(f'cachable:{hurl}', False)
        if cachable:
            print('caching', hurl)
            path = cache_root / f'{hurl}.opus'
            p = subprocess.Popen(
                [
                    'ffmpeg', '-hide_banner', '-loglevel', 'warning',
                    '-reconnect', '1', '-reconnect_at_eof', '0',
                    '-reconnect_streamed', '1', '-reconnect_delay_max', '10', '-copy_unknown',
                    '-y', '-i', rurl, str(path)
                ],
                # stdout=subprocess.PIPE,
                # stderr=subprocess.PIPE,
            )
            with Benchmark('CCH'):
                code = await loop.run_in_executor(None, p.wait)
            if code:
                raise RuntimeError(code)
            await cache_db.set(f'url:{hurl}', str(path))
            print('cached', hurl)
            # await cache_db.set(f'cachable:{hurl}', False)
        else:
            await cache_db.set(f'cachable:{hurl}', True)


async def real_url(url: str, override: bool) -> str:
    hurl: str = bytes_hash(url.encode())
    if not override:
        curl: Optional[str] = cache_db.get(f'url:{hurl}', None)
        if curl is not None:
            print('using cached', hurl)
            return curl
    p = subprocess.Popen(
        ['youtube-dl', '--no-playlist', '-f', 'bestaudio', '-g', '--', url],
        stdout=subprocess.PIPE
    )
    with Benchmark('URL'):
        code = await loop.run_in_executor(None, p.wait)
    if code:
        raise RuntimeError(code)
    rurl: str = p.stdout.readline().decode()[:-1]
    loop.create_task(cache_url(hurl, rurl, override))
    return rurl


async def create_ytaudio(ctx: Context, info: dict[str, Any], effects: Optional[str], already_read: int) -> YTAudio:
    if effects:
        permissions: discord.Permissions = ctx.member.guild_permissions
        if not permissions.administrator:
            raise Explicit("not an administrator")
        if not set(effects) <= set(string.ascii_letters + string.digits + '*,=+-/()|.^:_'):
            raise Explicit('malformed effects')
        options = f'-af {shlex.quote(effects)}'
    else:
        options = None
    return YTAudio(
        await real_url(info['url'], False),
        info['url'],
        f'{escape(info.get("title"))} `Rby` {ctx.member}',
        options,
        ctx.member,
        already_read
    )


async def entries_for_url(url: str) -> AsyncIterable[
    dict[str, Any]
]:
    info = await aextract(
        {
            "playlistend": 128,
            "logtostderr": True
        },
        url,
        download=False,
        process=False
    )
    if 'entries' in info:
        for entry in info['entries']:
            yield entry
    else:
        yield info | {'url': url}


async def create_ytaudios(ctx: Context, infos: list[tuple[dict[str, Any], str, int]]) -> AsyncIterable[YTAudio]:
    for audio in await asyncio.gather(
            *[
                create_ytaudio(ctx, info, effects, already_read)
                for
                info, effects, already_read
                in
                infos
            ]
    ):
        yield audio


presets: dict[str, str] = {
    'cursed': 'aeval=val(0)*2*sin(440*t)+val(1)*2*cos(622*t)|val(1)*2*sin(622*t)+val(0)*2*cos(440*t)',
    'bassboost': 'bass=g=10',
    'bassbooboost': 'bass=g=30',
    'nightcore': 'asetrate=67882',
    'пришествие анимешне': 'bass=g=15,asetrate=67882,bass=g=15',
    'difference': 'aeval=val(0)-val(1)|val(1)-val(0)',
    'mono': 'aeval=.5*val(0)+.5*val(1)|.5*val(1)+.5*val(0)',
}


async def entries_effects_for_args(args: list[str]) -> AsyncIterable[tuple[dict[str, Any], str, int]]:
    while args:
        match args:
            case [url, '-', effects, *args]:
                pass
            case [url, '+', preset, *args]:
                effects = presets[preset]
            case [url, *args]:
                effects = None
            case _:
                raise RuntimeError
        timestamp = 0
        match args:
            case [h, m, s, *args] if h.isdecimal() and m.isdecimal() and s.isdecimal():
                timestamp = 3600 * int(h) + 60 * int(m) + int(s)
            case [m, s, *args] if m.isdecimal() and s.isdecimal():
                timestamp = 60 * int(m) + int(s)
            case [s, *args] if s.isdecimal():
                timestamp = int(s)
            case [*args]:
                pass
        already_read = timestamp * 1000 / discord.opus.Encoder.FRAME_LENGTH
        async for info in entries_for_url(url):
            yield info, effects, already_read


async def yt_audios(ctx: Context, args: list[str]) -> AsyncIterable[YTAudio]:
    tuples: list[tuple[dict[str, Any], str, int]] = []
    async for info, effects, already_read in entries_effects_for_args(args):
        tuples.append((info, effects, already_read))
        if len(tuples) >= 5:
            async for audio in create_ytaudios(ctx, tuples):
                yield audio
            tuples.clear()
    async for audio in create_ytaudios(ctx, tuples):
        yield audio


mainasrcs: dict[discord.Guild, MainAudio] = {}


@at('commands', 'play')
async def play(ctx: Context, args: list[str]) -> None:
    match args:
        case [] | ['help']:
            await ctx.reply('''
`play ...args`
`play url [- effects] ...args`
'''.strip())
        case _:
            async with lock_for(ctx.guild):
                queue = await queue_for(ctx, create=True)
                async for audio in yt_audios(ctx, args):
                    queue.append(audio)
                await ctx.reply('done')


async def raw_vc_for(ctx: Context) -> discord.VoiceClient:
    if ctx.guild is None:
        raise Explicit("not in a guild")
    vc: discord.VoiceProtocol = ctx.guild.voice_client
    if vc is None or isinstance(vc, discord.VoiceClient) and not vc.is_connected():
        vs: discord.VoiceState = ctx.member.voice
        if vs is None:
            raise Explicit("not connected")
        vch: discord.VoiceChannel = vs.channel
        if vch is None:
            raise Explicit("not connected")
        try:
            vc: discord.VoiceProtocol = await vch.connect()
        except discord.ClientException:
            await ctx.guild.fetch_channels()
            raise Explicit("try again later")
    assert isinstance(vc, discord.VoiceClient)
    return vc


async def main_for_raw_vc(vc: discord.VoiceClient, *, create: bool) -> MainAudio:
    if vc.guild in mainasrcs:
        source = mainasrcs[vc.guild]
    else:
        if create:
            source = mainasrcs.setdefault(
                vc.guild,
                MainAudio(await QueueAudio.create(vc.guild), volume=volume_db.get(vc.guild.id, 0.2))
            )
        else:
            raise Explicit("not playing")
    if vc.source != source or not vc.is_playing():
        vc.play(source)
    return source


async def vc_main_for(ctx: Context, *, create: bool) -> tuple[discord.VoiceClient, MainAudio]:
    vc = await raw_vc_for(ctx)
    return vc, await main_for_raw_vc(vc, create=create)


async def vc_for(ctx: Context, *, create: bool) -> discord.VoiceClient:
    vc, source = await vc_main_for(ctx, create=create)
    return vc


async def main_for(ctx: Context, *, create: bool) -> MainAudio:
    vc, source = await vc_main_for(ctx, create=create)
    return source


async def queue_for(ctx: Context, *, create: bool) -> QueueAudio:
    return (await main_for(ctx, create=create)).queue


@at('commands', 'skip')
async def skip(ctx: Context, args: list[str]) -> None:
    match args:
        case ['help']:
            await ctx.reply('''
`skip [first] [last]`
'''.strip())
            return
        case []:
            queue = await queue_for(ctx, create=False)
            queue.skip_at(0, ctx.member)
        case [pos] if pos.isdecimal():
            pos = int(pos)
            queue = await queue_for(ctx, create=False)
            queue.skip_at(pos, ctx.member)
        case [pos0, pos1] if pos0.isdecimal() and pos1.isdecimal():
            pos0, pos1 = int(pos0), int(pos1)
            queue = await queue_for(ctx, create=False)
            for i in range(pos0, pos1 + 1):
                if not queue.skip_at(pos0, ctx.member):
                    pos0 += 1
        case _:
            raise Explicit("misformatted")
    await ctx.reply('done')


@at('commands', 'queue')
async def queue_(ctx: Context, args: list[str]) -> None:
    match args:
        case ['help']:
            await ctx.reply('current queue')
        case []:
            await ctx.long((await queue_for(ctx, create=False)).format().strip() or 'no queue')
        case ['clear']:
            (await queue_for(ctx, create=False)).clear(ctx.member)
            await ctx.reply('done')
        case ['resume']:
            async with lock_for(ctx.guild):
                await queue_for(ctx, create=True)
        case _:
            raise Explicit("misformatted")


@at('commands', 'swap')
async def swap(ctx: Context, args: list[str]) -> None:
    match args:
        case ['help']:
            await ctx.reply('`swap a b`')
        case [a, b] if a.isdecimal() and b.isdecimal():
            a, b = int(a), int(b)
            (await queue_for(ctx, create=False)).swap(ctx.member, a, b)
        case _:
            raise Explicit("misformatted")


@at('commands', 'move')
async def move(ctx: Context, args: list[str]) -> None:
    match args:
        case ['help']:
            await ctx.reply('`move a b`')
        case [a, b] if a.isdecimal() and b.isdecimal():
            a, b = int(a), int(b)
            (await queue_for(ctx, create=False)).move(ctx.member, a, b)
        case _:
            raise Explicit("misformatted")


@at('commands', 'volume')
async def volume_(ctx: Context, args: list[str]) -> None:
    match args:
        case ['help']:
            await ctx.reply('`volume 0.2`')
        case [volume]:
            volume = float(volume)
            await (await main_for(ctx, create=False)).set(volume, ctx.member)
        case _:
            raise Explicit("misformatted")


@at('commands', 'pause')
async def pause(ctx: Context, _args: list[str]) -> None:
    vc = await vc_for(ctx, create=False)
    vc.pause()


@at('commands', 'resume')
async def resume(ctx: Context, _args: list[str]) -> None:
    vc = await vc_for(ctx, create=False)
    vc.resume()


async def handle_args(message: discord.Message, args: list[str]):
    match args:
        case []:
            return
        case [command_name, *command_args]:
            ctx = Context(message)
            try:
                await handle_command(ctx, command_name, command_args)
            except Implicit:
                pass
            except Explicit as e:
                await ctx.reply(e.msg)


@client.event
async def on_message(message: discord.Message) -> None:
    if message.author.bot:
        return
    content: str = message.content
    if not content.startswith(prefix):
        return
    content = content.removeprefix(prefix)
    args = shlex.split(content)
    await handle_args(message, args)


async def save_queues():
    for mainasrc in list(mainasrcs.values()):
        await asyncio.sleep(0.01)
        await mainasrc.queue.save()


async def save_vcs():
    if vcs_restored:
        vcs = []
        vc: discord.VoiceClient
        for vc in list(client.voice_clients):
            await asyncio.sleep(0.01)
            if vc.is_playing():
                vcs.append((vc.guild.id, vc.channel.id, vc.is_paused()))
        queue_db.set_nowait('vcs', vcs)


async def save_commit():
    await queue_db.set('commit', time.time())


async def save_job():
    while True:
        await asyncio.sleep(1)
        with Benchmark('SVQ'):
            await save_queues()
        with Benchmark('SVV'):
            await save_vcs()
        with Benchmark('SVC'):
            await save_commit()


async def main():
    async with volume_db, queue_db, cache_db:
        await start_app(get_app(client))
        await client.login(token)
        loop.create_task(save_job())
        if os.getenv('v6monitor'):
            loop.create_task(monitor())
        await client.connect()


if __name__ == '__main__':
    loop.run_until_complete(main())
